'use strict';
var doop = require('jsdoc/util/doop');
var env = require('jsdoc/env');
var fs = require('jsdoc/fs');
var helper = require('jsdoc/util/templateHelper');
var logger = require('jsdoc/util/logger');
var path = require('jsdoc/path');
var taffy = require('taffydb').taffy;
var template = require('jsdoc/template');
var util = require('util');
var pct = require('./static/scripts/parse_complex_type');

var htmlsafe = helper.htmlsafe;
var linkto = pct.linkto;
var replace_enters_to_br = pct.replace_enters_to_br;


var resolveAuthorLinks = helper.resolveAuthorLinks;
var hasOwnProp = Object.prototype.hasOwnProperty;



var data;
var view;

console.log(env.opts.destination)
var outdir = path.normalize(env.opts.destination);
console.log(outdir)

function find(spec) {
    return helper.find(data, spec);
}

function tutoriallink(tutorial) {
    return helper.toTutorial(tutorial, null, {
        tag: 'em',
        classname: 'disabled',
        prefix: 'Tutorial: '
    });
}

function getAncestorLinks(doclet) {
    return helper.getAncestorLinks(data, doclet);
}

function hashToLink(doclet, hash) {
    var url;

    if ( !/^(#.+)/.test(hash) ) {
        return hash;
    }

    url = helper.createLink(doclet);
    url = url.replace(/(#.+|$)/, hash);

    return '<a href="' + url + '">' + hash + '</a>';
}

function needsSignature(doclet) {
    var needsSig = false;

    // function and class definitions always get a signature
    if (doclet.kind === 'function' || doclet.kind === 'class') {
        needsSig = true;
    }
    // typedefs that contain functions get a signature, too
    else if (doclet.kind === 'typedef' && doclet.type && doclet.type.names &&
        doclet.type.names.length) {
        for (var i = 0, l = doclet.type.names.length; i < l; i++) {
            if (doclet.type.names[i].toLowerCase() === 'function') {
                needsSig = true;
                break;
            }
        }
    }
    // and namespaces that are functions get a signature (but finding them is a
    // bit messy)
    else if (doclet.kind === 'namespace' && doclet.meta && doclet.meta.code &&
        doclet.meta.code.type && doclet.meta.code.type.match(/[Ff]unction/)) {
        needsSig = true;
    }

    return needsSig;
}

function getSignatureAttributes(item) {
    var attributes = [];

    if (item.optional) {
        attributes.push('opt');
    }

    if (item.nullable === true) {
        attributes.push('nullable');
    }
    else if (item.nullable === false) {
        attributes.push('non-null');
    }

    return attributes;
}

function updateItemName(item) {
    var attributes = getSignatureAttributes(item);
    var itemName = item.name || '';

    if (item.variable) {
        itemName = '&hellip;' + itemName;
    }

    if (attributes && attributes.length) {
        itemName = util.format( '%s<span class="signature-attributes">%s</span>', itemName,
            attributes.join(', ') );
    }

    return itemName;
}

function addParamAttributes(params) {
    return params.filter(function(param) {
        return param.name && param.name.indexOf('.') === -1;
    }).map(updateItemName);
}

function buildItemTypeStrings(item) {
    var types = [];

    if (item && item.type && item.type.names) {
        item.type.names.forEach(function(name) {
            types.push( linkto(name, htmlsafe(name)) );
        });
    }

    return types;
}

function buildAttribsString(attribs) {
    var attribsString = '';

    if (attribs && attribs.length) {
        attribsString = htmlsafe( util.format('(%s) ', attribs.join(', ')) );
    }

    return attribsString;
}

function addNonParamAttributes(items) {
    var types = [];

    items.forEach(function(item) {
        types = types.concat( buildItemTypeStrings(item) );
    });

    return types;
}

function addSignatureParams(f) {
    var params = f.params ? addParamAttributes(f.params) : [];

    f.signature = util.format( '%s(%s)', (f.signature || ''), params.join(', ') );
}

function addSignatureReturns(f) {
    var attribs = [];
    var attribsString = '';
    var returnTypes = [];
    var returnTypesString = '';
    var source = f.yields || f.returns;

    // jam all the return-type attributes into an array. this could create odd results (for example,
    // if there are both nullable and non-nullable return types), but let's assume that most people
    // who use multiple @return tags aren't using Closure Compiler type annotations, and vice-versa.
    if (source) {
        source.forEach(function(item) {
            helper.getAttribs(item).forEach(function(attrib) {
                if (attribs.indexOf(attrib) === -1) {
                    attribs.push(attrib);
                }
            });
        });

        attribsString = buildAttribsString(attribs);
    }

    if (source) {
        returnTypes = addNonParamAttributes(source);
    }
    if (returnTypes.length) {
        returnTypesString = util.format( ' &rarr; %s{%s}', attribsString, returnTypes.join('|') );
    }

    f.signature = '<span class="signature">' + (f.signature || '') + '</span>' +
        '<span class="type-signature">' + returnTypesString + '</span>';
}

function addSignatureTypes(f) {
    var types = f.type ? buildItemTypeStrings(f) : [];

    f.signature = (f.signature || '') + '<span class="type-signature">' +
        (types.length ? ' :' + types.join('|') : '') + '</span>';
}

function addAttribs(f) {
    var attribs = helper.getAttribs(f);
    var attribsString = buildAttribsString(attribs);

    f.attribs = util.format('<span class="type-signature">%s</span>', attribsString);
}

function shortenPaths(files, commonPrefix) {
    Object.keys(files).forEach(function(file) {
        files[file].shortened = files[file].resolved.replace(commonPrefix, '')
            // always use forward slashes
            .replace(/\\/g, '/');
    });

    return files;
}

function getPathFromDoclet(doclet) {
    if (!doclet.meta) {
        return null;
    }

    return doclet.meta.path && doclet.meta.path !== 'null' ?
        path.join(doclet.meta.path, doclet.meta.filename) :
        doclet.meta.filename;
}

function generate(title, docs, filename, resolveLinks) {
    var docData;
    var html;
    var outpath;

    resolveLinks = resolveLinks !== false;

    docData = {
        env: env,
        title: title,
        docs: docs
    };

    outpath = path.join(outdir, filename);
    // console.log("outdir:", outdir, "outpath:", outpath, "filename:", filename)

    // console.log(docData)
    html = view.render('container.tmpl', docData);

    if (resolveLinks) {
        html = helper.resolveLinks(html); // turn {@link foo} into <a href="foodoc.html">foo</a>
    }

    fs.writeFileSync(outpath, html, 'utf8');
    return outpath;
}

function generateSourceFiles(sourceFiles, encoding) {
    encoding = encoding || 'utf8';
    Object.keys(sourceFiles).forEach(function(file) {
        var source;
        // links are keyed to the shortened path in each doclet's `meta.shortpath` property
        var sourceOutfile = helper.getUniqueFilename(sourceFiles[file].shortened);

        helper.registerLink(sourceFiles[file].shortened, sourceOutfile);

        try {
            source = {
                kind: 'source',
                code: helper.htmlsafe( fs.readFileSync(sourceFiles[file].resolved, encoding) )
            };
        }
        catch (e) {
            logger.error('Error while generating source file %s: %s', file, e.message);
        }

        generate('Source: ' + sourceFiles[file].shortened, [source], sourceOutfile,
            false);
    });
}

/**
 * Look for classes or functions with the same name as modules (which indicates that the module
 * exports only that class or function), then attach the classes or functions to the `module`
 * property of the appropriate module doclets. The name of each class or function is also updated
 * for display purposes. This function mutates the original arrays.
 *
 * @private
 * @param {Array.<module:jsdoc/doclet.Doclet>} doclets - The array of classes and functions to
 * check.
 * @param {Array.<module:jsdoc/doclet.Doclet>} modules - The array of module doclets to search.
 */
function attachModuleSymbols(doclets, modules) {
    var symbols = {};

    // build a lookup table
    doclets.forEach(function(symbol) {
        symbols[symbol.longname] = symbols[symbol.longname] || [];
        symbols[symbol.longname].push(symbol);
    });

    modules.forEach(function(module) {
        if (symbols[module.longname]) {
            module.modules = symbols[module.longname]
                // Only show symbols that have a description. Make an exception for classes, because
                // we want to show the constructor-signature heading no matter what.
                .filter(function(symbol) {
                    return symbol.description || symbol.kind === 'class';
                })
                .map(function(symbol) {
                    symbol = doop(symbol);

                    if (symbol.kind === 'class' || symbol.kind === 'function') {
                        symbol.name = symbol.name.replace('module:', '(require("') + '"))';
                    }

                    return symbol;
                });
        }
    });
}

function buildMemberNav(items, itemHeading, itemsSeen, linktoFn) {
    var nav = '';

    if (items.length) {
        var itemsNav = '';
        items.forEach(function(item) {
            var displayName;

            if ( !hasOwnProp.call(item, 'longname') ) {
                itemsNav += '<li>' + linktoFn('', item.name) + '</li>';
            }
            else if ( !hasOwnProp.call(itemsSeen, item.longname) ) {
                if (env.conf.templates.default.useLongnameInNav) {
                    displayName = item.longname;
                } else {
                    displayName = item.longname;
                }
                itemsNav += '<li>' + linktoFn(item.longname, displayName.replace(/\b(module|event):/g, '')) + '</li>';

                itemsSeen[item.longname] = true;
            }
        });

        if (itemsNav !== '') {
            nav += '<h3>' + itemHeading + '</h3><ul>' + itemsNav + '</ul>';
        }
    }

    return nav;
}




function buildMemberNavNamespacesTree(items, itemHeading, itemsSeen, linktoFn) {
    var nav = '';

    var restored_items = restoreSourceTree(items);

    var recursive_generate_namespaces_old = function (_tree) {
        var items = _tree.items;
        var a = 0;
        var res = "";
        if (_tree.item_data) res = "<ul class=\"sub-menu\" id=\"" + _tree.item_data.name + "\">";
        if (_tree.item_data) res += "<li>" + linktoFn(_tree.item_data.longname, _tree.item_data.name.replace(/\b(module|event):/g, '')) + "</li>";
        while (a < items.length) {
            var name = items[a];
            var data = _tree[name];
            res += recursive_generate_namespaces(data);
            a++;
        }
        if (_tree.item_data) res += "</ul>";
        return res;
    };

    var g_id = 0;
    var recursive_generate_namespaces = function (_tree) {
        var lgid = g_id++;
        var items = _tree.items;
        var a = 0;
        var res = "";
        if (_tree.item_data) res = "<li>";
        if (_tree.item_data) res += "<input type=\"radio\" class=\"mostrar-menu\" id=\"menu" + lgid + "\">";
        if (items.length != 0 && _tree.item_data) res += "<label for=\"menu" + lgid + "\" class=\"ampliar\"></label>";
        if (_tree.item_data) res += linktoFn(_tree.item_data.longname, _tree.item_data.name.replace(/\b(module|event):/g, ''));

        if (items.length != 0) {
            if (_tree.item_data) res += "<ul>";
            while (a < items.length) {
                var name = items[a];
                var data = _tree[name];
                res += recursive_generate_namespaces(data);
                a++;
            }
            if (_tree.item_data) res += "</ul>";
        }
        if (_tree.item_data) res += "</li>";

        return res;
    };

    var req_res = recursive_generate_namespaces(restored_items);

    debugger;
    if (items.length) {
        var itemsNav = '';

        itemsNav += req_res;

        if (itemsNav !== '') {
            nav += '<h3>' + itemHeading + '</h3><ul class="menu-arbol">' + itemsNav + '</ul>';
        }
    }

    return nav;
}
function restoreSourceTree(items) {
    var out_tree = {items:[]};
    var a = 0;
    while( a < items.length){
        var item = items[a];
        var path = item.longname.split(".");
        // console.log(item.longname);
        var cur_path = out_tree;
        var b = 0;
        while(b < path.length){
            var hop = path[b];
            var is_exist = !!cur_path[hop];
            // console.log("current hop: " + hop);
            if(!is_exist){
                // console.log("new hop: " + hop);
                cur_path[hop] = {};
                cur_path[hop].item_data = item;
                cur_path[hop].items = [];
                cur_path.items.push(hop)
            } else {
                // console.log("exist hop: " + hop);
                cur_path = cur_path[hop];
            }
            b++;
        }
        // console.log("------------");
        a++;
    }
    return out_tree;
}

function linktoTutorial(longName, name) {
    return tutoriallink(name);
}

function linktoExternal(longName, name) {
    return linkto(longName, name.replace(/(^"|"$)/g, ''));
}

/**
 * Create the navigation sidebar.
 * @param {object} members The members that will be used to create the sidebar.
 * @param {array<object>} members.classes
 * @param {array<object>} members.externals
 * @param {array<object>} members.globals
 * @param {array<object>} members.mixins
 * @param {array<object>} members.modules
 * @param {array<object>} members.namespaces
 * @param {array<object>} members.tutorials
 * @param {array<object>} members.events
 * @param {array<object>} members.interfaces
 * @return {string} The HTML for the navigation sidebar.
 */
// function buildNav(members) {
//     var globalNav;
//     var nav = '<h2><a href="index.html">Home</a></h2>';
//     var seen = {};
//     var seenTutorials = {};
//
//     nav += buildMemberNav(members.modules, 'Modules', {}, linkto);
//     nav += buildMemberNav(members.externals, 'Externals', seen, linktoExternal);
//     nav += buildMemberNavNamespace(members.namespaces, 'Managers', seen, linkto);
//     nav += buildMemberNavNamespace(members.classes, 'Classes', seen, linkto);
//     nav += buildMemberNav(members.events, 'Events', seen, linkto);
//     // nav += buildMemberNavNamespaces(members.namespaces, 'Namespaces', seen, linkto);
//     // nav += buildMemberNav(members.namespaces, 'Namespaces', seen, linkto);
//     nav += buildMemberNav(members.mixins, 'Mixins', seen, linkto);
//     nav += buildMemberNav(members.tutorials, 'Tutorials', seenTutorials, linktoTutorial);
//     nav += buildMemberNav(members.interfaces, 'Interfaces', seen, linkto);
//
//     if (members.globals.length) {
//         globalNav = '';
//
//         members.globals.forEach(function(g) {
//             if ( g.kind !== 'typedef' && !hasOwnProp.call(seen, g.longname) ) {
//                 globalNav += '<li>' + linkto(g.longname, g.name) + '</li>';
//             }
//             seen[g.longname] = true;
//         });
//
//         if (!globalNav) {
//             // turn the heading into a link so you can actually get to the global page
//             nav += '<h3>' + linkto('global', 'Global') + '</h3>';
//         }
//         else {
//             nav += '<h3>Global</h3><ul>' + globalNav + '</ul>';
//         }
//     }
//
//     return nav;
// }

function buildMemberNavNamespace(items, itemHeading, itemsSeen, linktoFn) {
    var nav = '';

    console.log("buildMemberNavNamespace");
    console.log("items", JSON.stringify(items, true, 3));
    console.log("itemHeading", JSON.stringify(itemHeading, true, 3));
    console.log("itemsSeen", JSON.stringify(itemsSeen, true, 3));
    console.log("linktoFn", JSON.stringify(linktoFn, true, 3));

    if (items.length) {
        var itemsNav = '';
        items.forEach(function(item) {
            var displayName;

            if ( !hasOwnProp.call(item, 'longname') ) {
                itemsNav += '<li>' + linktoFn('', item.name) + '</li>';
            }
            else if ( !hasOwnProp.call(itemsSeen, item.longname) ) {
                displayName = item.longname;

                var arr = displayName.split(".");
                var end_id = arr.length - 1;
                var last = arr.pop();

                var prev_name = arr.join(".");
                var res = "";
                // res += "<span title='"+item.comment+"'>";
                if (arr.length > 0) res += "<span class='__prev'>" +prev_name+ ".</span>";
                res += "<span class='__end'>" +last+ "</span>";
                itemsNav += '<li>' + linktoFn(item.longname, res.replace(/\b(module|event):/g, '')) + '</li>';
                // itemsNav += "</span>";
                itemsSeen[item.longname] = true;
            }
        });

        if (itemsNav !== '') {
            nav += '<h3>' + itemHeading + '</h3><ul>' + itemsNav + '</ul>';
        }
    }

    return nav;
}

function buildNav(members) {
    var nav = '<h2 class="ddm"><a href="index.html">GUI 2.9.1-develop</a></h2>';

    nav += render_explorer(members.namespaces, members.classes);

    return nav;
}

var render_explorer = function (_nss, _clss) {
    var nav = "";
    // nav += render_menu_item("Widgets", "widgets", _nss, _clss, true);
    // nav += render_menu_item("Datasources", "datasources", _nss, _clss, true);
    // nav += render_menu_item("SignalSlot", "sigslot", _nss, _clss, true);
    // nav += render_menu_item("Grid", "grid", _nss, _clss, true);
    nav += render_menu_item("Providers Interface", "providers_interface", _nss, _clss, true);
    nav += render_menu_item("Providers Managers", "manager", _nss, _clss, true);
    nav += render_menu_item("Libraries", "library", _nss, _clss, true);
    return nav;
};


var render_menu_item = function (_title, _type, _nss, _clss, _check_tags) {
    var nav = "";

    nav += "<ul class='" + _title + " ddm'>";
    nav += "<div class='title'>" + _title + "</div>";

    for (var a = 0; a < _nss.length; a++) {
        var ns = _nss[a];

        var is_manager = _check_tags && ns.tags && check_tags(ns.tags, _type);

        if (!is_manager) continue;

        var members = find_first_members(ns.name, _clss);

        nav += "<li class='dropdown'>";

        if (members.length > 0) {
            nav += linkto(ns.longname, ns.name + "<i class=\"icon-arrow\"></i>", "menu-title");
        } else {
            nav += linkto(ns.longname, ns.name, "menu-title");
        }

        if (members.length > 0) {
            nav += "<ul id='" + ns.name + "' class='dropdown-menu'>";
            for (var b = 0; b < members.length; b++) {
                var member = members[b];
                var li_start = "<li class=\"" + _title + "\">";
                var link = linkto(member.longname, member.name);
                var li_end = "</li>";
                nav += li_start + link + li_end;
            }
            nav += "</ul>";
        }
        nav += "</li>";
    }
    nav += "</ul>";

    return nav;
};

var check_tags = function (_tags, _key) {
    for (var a = 0; a < _tags.length; a++) {
        var info = _tags[a];
        if (info.title == "group" && info.text == _key) return true;
    }
    return false;
};

var find_first_members = function (_parent, _clss) {
    var members = [];
    for(var a = 0; a < _clss.length; a++ ) {
        var cls = _clss[a];

        var hierarchy = cls.memberof.split(".");
        if (hierarchy.length == 1 && hierarchy[0] === "") {
            continue;
        }

        var found = false;
        for(var b = 0; b < hierarchy.length; b++ ) {
            var hop = hierarchy[b];
            if(hop == _parent) {
                found = true;
                break;
            }
        }

        var member_success = found && hierarchy.length - b < 2;

        if(!member_success) continue;

        members.push(cls);
    }

    return members;
};

global.typedefs = {};
global.mainpage_content = [];
global.cusomData = Object.create(null);

var find_tag_by_id = function (_tags, _tag_id) {
    for (var a = 0; a < _tags.length; a++) {
        var info = _tags[a];
        if (info.title == _tag_id) return true;
    }
    return false;
};

/**
    @param {TAFFY} taffyData See <http://taffydb.com/>.
    @param {object} opts
    @param {Tutorial} tutorials
 */
exports.publish = function(taffyData, opts, tutorials) {
    var classes;
    var conf;
    var externals;
    var files;
    var fromDir;
    var globalUrl;
    var indexUrl;
    var interfaces;
    var members;
    var mixins;
    var modules;
    var namespaces;
    var outputSourceFiles;
    var packageInfo;
    var packages;
    var sourceFilePaths = [];
    var sourceFiles = {};
    var staticFileFilter;
    var staticFilePaths;
    var staticFiles;
    var staticFileScanner;
    var templatePath;

    data = taffyData;

    conf = env.conf.templates || {};
    conf.default = conf.default || {};

    templatePath = path.normalize(opts.template);
    view = new template.Template( path.join(templatePath, 'tmpl') );

    // claim some special filenames in advance, so the All-Powerful Overseer of Filename Uniqueness
    // doesn't try to hand them out later
    indexUrl = helper.getUniqueFilename('index');
    // don't call registerLink() on this one! 'index' is also a valid longname

    globalUrl = helper.getUniqueFilename('global');
    helper.registerLink('global', globalUrl);

    // set up templating
    view.layout = conf.default.layoutFile ?
        path.getResourcePath(path.dirname(conf.default.layoutFile),
            path.basename(conf.default.layoutFile) ) :
        'layout.tmpl';

    // set up tutorials for helper
    helper.setTutorials(tutorials);

    data = helper.prune(data);
    data.sort('longname, version, since');
    helper.addEventListeners(data);

    data().each(function(doclet){
        if(doclet.kind == "typedef") {
            typedefs[doclet.longname] = doclet;
        }

        if(doclet.tags && find_tag_by_id(doclet.tags, "mainpage")){
            mainpage_content.push(doclet);
        }


        if(doclet.meta && doclet.meta.path) {
            var path = doclet.meta.path.split("/");
            if(path[path.length - 1] === "custom") {
                var id = doclet.meta.filename.split(".")[0];

                if (!cusomData[id])
                    cusomData[id] = [];

                if (doclet.tags && find_tag_by_id(doclet.tags, id))
                    cusomData[id].push(doclet);
            }
        }
    });


    data().each(function(doclet) {

        var sourcePath;

        doclet.attribs = '';

        if (doclet.examples) {
            doclet.examples = doclet.examples.map(function(example) {
                var caption;
                var code;

                if (example.match(/^\s*<caption>([\s\S]+?)<\/caption>(\s*[\n\r])([\s\S]+)$/i)) {
                    caption = RegExp.$1;
                    code = RegExp.$3;
                }

                return {
                    caption: caption || '',
                    code: code || example
                };
            });
        }
        if (doclet.see) {
            doclet.see.forEach(function(seeItem, i) {
                doclet.see[i] = hashToLink(doclet, seeItem);
            });
        }

        // build a list of source files
        if (doclet.meta) {
            sourcePath = getPathFromDoclet(doclet);
            sourceFiles[sourcePath] = {
                resolved: sourcePath,
                shortened: null
            };
            if (sourceFilePaths.indexOf(sourcePath) === -1) {
                sourceFilePaths.push(sourcePath);
            }
        }
    });

    // update outdir if necessary, then create outdir
    packageInfo = ( find({kind: 'package'}) || [] )[0];
    if (packageInfo && packageInfo.name) {
        outdir = path.join( outdir, packageInfo.name, (packageInfo.version || '') );
    }
    fs.mkPath(outdir);

    // copy the template's static files to outdir
    fromDir = path.join(templatePath, 'static');
    staticFiles = fs.ls(fromDir, 3);

    staticFiles.forEach(function(fileName) {
        var toDir = fs.toDir( fileName.replace(fromDir, outdir) );

        fs.mkPath(toDir);
        fs.copyFileSync(fileName, toDir);
    });

    // copy user-specified static files to outdir
    if (conf.default.staticFiles) {
        // The canonical property name is `include`. We accept `paths` for backwards compatibility
        // with a bug in JSDoc 3.2.x.
        staticFilePaths = conf.default.staticFiles.include ||
            conf.default.staticFiles.paths ||
            [];
        staticFileFilter = new (require('jsdoc/src/filter')).Filter(conf.default.staticFiles);
        staticFileScanner = new (require('jsdoc/src/scanner')).Scanner();

        staticFilePaths.forEach(function(filePath) {
            var extraStaticFiles;

            filePath = path.resolve(env.pwd, filePath);
            extraStaticFiles = staticFileScanner.scan([filePath], 10, staticFileFilter);

            extraStaticFiles.forEach(function(fileName) {
                var sourcePath = fs.toDir(filePath);
                var toDir = fs.toDir( fileName.replace(sourcePath, outdir) );

                fs.mkPath(toDir);
                fs.copyFileSync(fileName, toDir);
            });
        });
    }

    if (sourceFilePaths.length) {
        sourceFiles = shortenPaths( sourceFiles, path.commonPrefix(sourceFilePaths) );
    }
    data().each(function(doclet) {
        var docletPath;
        var url = helper.createLink(doclet);

        helper.registerLink(doclet.longname, url);

        // add a shortened version of the full path
        if (doclet.meta) {
            docletPath = getPathFromDoclet(doclet);
            docletPath = sourceFiles[docletPath].shortened;
            if (docletPath) {
                doclet.meta.shortpath = docletPath;
            }
        }
    });

    data().each(function(doclet) {
        var url = helper.longnameToUrl[doclet.longname];

        if (url.indexOf('#') > -1) {
            doclet.id = helper.longnameToUrl[doclet.longname].split(/#/).pop();
        }
        else {
            doclet.id = doclet.name;
        }

        if ( needsSignature(doclet) ) {
            addSignatureParams(doclet);
            addSignatureReturns(doclet);
            addAttribs(doclet);
        }
    });

    // do this after the urls have all been generated
    data().each(function(doclet) {
        doclet.ancestors = getAncestorLinks(doclet);

        if (doclet.kind === 'member') {
            addSignatureTypes(doclet);
            addAttribs(doclet);
        }

        if (doclet.kind === 'constant') {
            addSignatureTypes(doclet);
            addAttribs(doclet);
            doclet.kind = 'member';
        }
    });

    members = helper.getMembers(data);
    members.tutorials = tutorials.children;

    // output pretty-printed source files by default
    outputSourceFiles = conf.default && conf.default.outputSourceFiles !== false;

    // add template helpers
    view.content = mainpage_content;
    view.cusomData = cusomData;
    view.find = find;
    view.linkto = linkto;
    view.replace_enters_to_br = replace_enters_to_br;
    view.resolveAuthorLinks = resolveAuthorLinks;
    view.tutoriallink = tutoriallink;
    view.htmlsafe = htmlsafe;
    view.outputSourceFiles = outputSourceFiles;

    // once for all
    view.nav = buildNav(members);
    attachModuleSymbols( find({ longname: {left: 'module:'} }), members.modules );

    // generate the pretty-printed source files first so other pages can link to them
    if (outputSourceFiles) {
        generateSourceFiles(sourceFiles, opts.encoding);
    }

    if (members.globals.length) { generate('Global', [{kind: 'globalobj'}], globalUrl); }

    // index page displays information from package.json and lists files
    files = find({kind: 'file'});
    packages = find({kind: 'package'});

    generate('Documentation: GUI 2.9.1-develop',
        packages.concat(
            [{
                kind: 'mainpage',
                readme: opts.readme,
                longname: (opts.mainpagetitle) ? opts.mainpagetitle : 'Main Page'
            }]
        ).concat(files), indexUrl);

    console.log("outdir", outdir)
    // console.log("indexUrl", indexUrl)



    for(var id in cusomData) {
        var out = generate(id,
            packages.concat(
                [{
                    subkind: id,
                    kind: 'custom',
                    readme: opts.readme,
                    longname: id
                }]
            ).concat(files), id + ".html");
        console.log("generate", id, out)

    }


    var sih = require("./search_inheritance.js");
    var inh_struct = sih.find(env.conf, helper, logger, generate);

    var sih2 = require("./search_inheritance2.js");
    var inh_struct2 = sih2.find(env.conf);

    var sih_core = require("./search_inheritance_core.js");
    var inh_struct_core = sih_core.find(env.conf);



    // Arch
    var html = view.render('arch.tmpl', {
        env: env,
        title: "",
    });
    html = helper.resolveLinks(html); // turn {@link foo} into <a href="foodoc.html">foo</a>
    // fs.createFileSync("/arch.html", html, 'utf8');
    fs.writeFileSync(outdir + "/arch.html", html, 'utf8');


    // providers tree
    var html = view.render('inheritance.tmpl', {
        inheritance: inh_struct,
        env: env,
        title: "",
    });
    html = helper.resolveLinks(html); // turn {@link foo} into <a href="foodoc.html">foo</a>
    // fs.createFileSync("/arch.html", html, 'utf8');
    fs.writeFileSync(outdir + "/inheritance.html", html, 'utf8');

    // gui tree
    html = view.render('inheritance2.tmpl', {
        inheritance: inh_struct2,
        height: 15100,
        env: env,
        title: "",
    });
    html = helper.resolveLinks(html); // turn {@link foo} into <a href="foodoc.html">foo</a>
    // fs.createFileSync("/arch.html", html, 'utf8');
    fs.writeFileSync(outdir + "/gui_tree.html", html, 'utf8');

    // gui graph
    html = view.render('inheritance3.tmpl', {
        inheritance: inh_struct2,
        height: 1600,
        env: env,
        title: "",
    });
    html = helper.resolveLinks(html); // turn {@link foo} into <a href="foodoc.html">foo</a>
    // fs.createFileSync("/arch.html", html, 'utf8');
    fs.writeFileSync(outdir + "/gui_graph.html", html, 'utf8');


    // core tree
    html = view.render('inheritance2.tmpl', {
        inheritance: inh_struct_core,
        height: 10000,
        env: env,
        title: "",
    });
    html = helper.resolveLinks(html);
    fs.writeFileSync(outdir + "/core_tree.html", html, 'utf8');


    // core graph
    html = view.render('inheritance3.tmpl', {
        inheritance: inh_struct_core,
        height: 800,
        env: env,
        title: "",
    });
    html = helper.resolveLinks(html);
    fs.writeFileSync(outdir + "/core_graph.html", html, 'utf8');


    // set up the lists that we'll use to generate pages
    classes = taffy(members.classes);
    modules = taffy(members.modules);
    namespaces = taffy(members.namespaces);
    mixins = taffy(members.mixins);
    externals = taffy(members.externals);
    interfaces = taffy(members.interfaces);

    Object.keys(helper.longnameToUrl).forEach(function(longname) {
        var myClasses = helper.find(classes, {longname: longname});
        var myExternals = helper.find(externals, {longname: longname});
        var myInterfaces = helper.find(interfaces, {longname: longname});
        var myMixins = helper.find(mixins, {longname: longname});
        var myModules = helper.find(modules, {longname: longname});
        var myNamespaces = helper.find(namespaces, {longname: longname});

        if (myModules.length) {
            generate('Module: ' + myModules[0].name, myModules, helper.longnameToUrl[longname]);
        }

        if (myClasses.length) {
            generate(myClasses[0].name, myClasses, helper.longnameToUrl[longname]);
        }

        if (myNamespaces.length) {
            generate(myNamespaces[0].name, myNamespaces, helper.longnameToUrl[longname]);
        }

        if (myMixins.length) {
            generate('Mixin: ' + myMixins[0].name, myMixins, helper.longnameToUrl[longname]);
        }

        if (myExternals.length) {
            generate('External: ' + myExternals[0].name, myExternals, helper.longnameToUrl[longname]);
        }

        if (myInterfaces.length) {
            generate('Interface: ' + myInterfaces[0].name, myInterfaces, helper.longnameToUrl[longname]);
        }
    });

    // TODO: move the tutorial functions to templateHelper.js
    function generateTutorial(title, tutorial, filename) {
        var tutorialData = {
            title: title,
            header: tutorial.title,
            content: tutorial.parse(),
            children: tutorial.children
        };
        var tutorialPath = path.join(outdir, filename);
        var html = view.render('tutorial.tmpl', tutorialData);

        // yes, you can use {@link} in tutorials too!
        html = helper.resolveLinks(html); // turn {@link foo} into <a href="foodoc.html">foo</a>

        fs.writeFileSync(tutorialPath, html, 'utf8');
    }

    // tutorials can have only one parent so there is no risk for loops
    function saveChildren(node) {
        node.children.forEach(function(child) {
            generateTutorial('Tutorial: ' + child.title, child, helper.tutorialToUrl(child.name));
            saveChildren(child);
        });
    }

    saveChildren(tutorials);
};
